@@ -63,6 +63,22 @@ public:
     static KGAPI::Object* JSONToEvent(const QVariantMap& event);
     static QVariantMap eventToJSON(KGAPI::Object *event);
     static QList<KGAPI::Object*> parseEventJSONFeed(const QVariantList& feed);
+
+    /**
+     * Checks whether TZID is in Olson format and converts it to it if neccessary
+     *
+     * This is mainly to handle crazy Microsoft TZIDs like
+     * "(GMT) Greenwich Mean Time/Dublin/Edinburgh/London", because Google only
+     * accepts TZIDs in Olson format ("Europe/London").
+     *
+     * It first tries to match the given \p tzid to all TZIDs in KTimeZones::zones().
+     * If it fails, it parses the \p event, looking for X-MICROSOFT-CDO-TZID
+     * property and than matches it to Olson-formatted TZID using a table.
+     *
+     * When the method fails to process the TZID, it returns the original \p tzid
+     * in hope, that Google will cope with it.
+     */
+    static QString checkAndConverCDOTZID(const QString &tzid, const Objects::Event *event);
 };
 
 }
@@ -627,7 +643,7 @@ QVariantMap Services::CalendarPrivate::eventToJSON(KGAPI::Object* event)
         tzStart = KTimeZone::utc().name();
     }
     if (!tzStart.isEmpty()) {
-        start["timeZone"] = tzStart;
+        start["timeZone"] = checkAndConverCDOTZID(tzStart, object);
     }
     data["start"] = start;
 
@@ -646,7 +662,7 @@ QVariantMap Services::CalendarPrivate::eventToJSON(KGAPI::Object* event)
         tzEnd = KTimeZone::utc().name();
     }
     if (!tzEnd.isEmpty()) {
-        end["timeZone"] = tzEnd;
+        end["timeZone"] = checkAndConverCDOTZID(tzEnd, object);
     }
     data["end"] = end;
 
@@ -794,3 +810,176 @@ DateList Services::CalendarPrivate::parseRDate(const QString& rule)
 
     return list;
 }
+
+
+static QMap<int, QString> initMSCDOTZIDTable()
+{
+    QMap<int, QString> map;
+
+    /* Based on "Time Zone to CdoTimeZoneId Map"
+     * http://msdn.microsoft.com/en-us/library/aa563018%28loband%29.aspx
+     *
+     * The mapping is not exact, since the CdoTimeZoneId usually refers to a
+     * region of multiple countries, so I always picked one of the countries
+     * in the specified region and used it's TZID.
+     */
+    map.insert(0,  QLatin1String("UTC"));
+    map.insert(1,  QLatin1String("Europe/London"));                       /* GMT Greenwich Mean Time; Dublin, Edinburgh, London */
+    /* Seriously? *sigh* Let's handle these two in checkAndConvertCDOTZID() */
+    //map.insertMulti(2, QLatin1String("Europe/Lisbon"));         /* GMT Greenwich Mean Time: Dublin, Edinburgh, Lisbon, London */
+    //map.insertMulti(2, QLatin1String("Europe/Sarajevo"));       /* GMT+01:00 Sarajevo, Skopje, Sofija, Vilnius, Warsaw, Zagreb */
+    map.insert(3,  QLatin1String("Europe/Paris"));              /* GMT+01:00 Paris, Madrid, Brussels, Copenhagen */
+    map.insert(4,  QLatin1String("Europe/Berlin"));             /* GMT+01:00 Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna */
+    map.insert(5,  QLatin1String("Europe/Bucharest"));          /* GMT+02:00 Bucharest */
+    map.insert(6,  QLatin1String("Europe/Prague"));             /* GMT+01:00 Prague, Central Europe */
+    map.insert(7,  QLatin1String("Europe/Athens"));             /* GMT+02:00 Athens, Istanbul, Minsk */
+    map.insert(8,  QLatin1String("America/Brazil"));            /* GMT-03:00 Brasilia */
+    map.insert(9,  QLatin1String("America/Halifax"));           /* GMT-04:00 Atlantic time (Canada) */
+    map.insert(10, QLatin1String("America/New_York"));          /* GMT-05:00 Eastern Time (US & Canada) */
+    map.insert(11, QLatin1String("America/Chicago"));           /* GMT-06:00 Central Time (US & Canada) */
+    map.insert(12, QLatin1String("America/Denver"));            /* GMT-07:00 Mountain Time (US & Canada) */
+    map.insert(13, QLatin1String("America/Los_Angeles"));       /* GMT-08:00 Pacific Time (US & Canada); Tijuana */
+    map.insert(14, QLatin1String("America/Anchorage"));         /* GMT-09:00 Alaska */
+    map.insert(15, QLatin1String("Pacific/Honolulu"));          /* GMT-10:00 Hawaii */
+    map.insert(16, QLatin1String("Pacific/Apia"));              /* GMT-11:00 Midway Island, Samoa */
+    map.insert(17, QLatin1String("Pacific/Auckland"));          /* GMT+12:00 Auckland, Wellington */
+    map.insert(18, QLatin1String("Australia/Brisbane"));        /* GMT+10:00 Brisbane, East Australia */
+    map.insert(19, QLatin1String("Australia/Adelaide"));        /* GMT+09:30 Adelaide, Central Australia */
+    map.insert(20, QLatin1String("Asia/Tokyo"));                /* GMT+09:00 Osaka, Sapporo, Tokyo */
+    map.insert(21, QLatin1String("Asia/Singapore"));            /* GMT+08:00 Kuala Lumpur, Singapore */
+    map.insert(22, QLatin1String("Asia/Bangkok"));              /* GMT+07:00 Bangkok, Hanoi, Jakarta */
+    map.insert(23, QLatin1String("Asia/Calcutta"));             /* GMT+05:30 Kolkata, Chennai, Mumbai, New Delhi, India Standard Time */
+    map.insert(24, QLatin1String("Asia/Dubai"));                /* GMT+04:00 Abu Dhabi, Muscat */
+    map.insert(25, QLatin1String("Asia/Tehran"));               /* GMT+03:30 Tehran */
+    map.insert(26, QLatin1String("Asia/Baghdad"));              /* GMT+03:00 Baghdad */
+    map.insert(27, QLatin1String("Asia/Jerusalem"));            /* GMT+02:00 Israel, Jerusalem Standard Time */
+    map.insert(28, QLatin1String("America/St_Johns"));          /* GMT-03:30 Newfoundland */
+    map.insert(29, QLatin1String("Atlantic/Portugal"));         /* GMT-01:00 Azores */
+    map.insert(30, QLatin1String("America/Noronha"));           /* GMT-02:00 Mid-Atlantic */
+    map.insert(31, QLatin1String("Africa/Monrovia"));           /* GMT Casablanca, Monrovia */
+    map.insert(32, QLatin1String("America/Argentina/Buenos_Aires")); /* GMT-03:00 Buenos Aires, Georgetown */
+    map.insert(33, QLatin1String("America/La_Paz"));            /* GMT-04:00 Caracas, La Paz */
+    map.insert(34, QLatin1String("America/New_York"));          /* GMT-05:00 Indiana (East) */
+    map.insert(35, QLatin1String("America/Bogota"));            /* GMT-05:00 Bogota, Lima, Quito */
+    map.insert(36, QLatin1String("America/Winnipeg"));          /* GMT-06:00 Saskatchewan */
+    map.insert(37, QLatin1String("America/Mexico_City"));       /* GMT-06:00 Mexico City, Tegucigalpa */
+    map.insert(38, QLatin1String("America/Phoenix"));           /* GMT-07:00 Arizona */
+    map.insert(39, QLatin1String("Pacific/Kwajalein"));         /* GMT-12:00 Eniwetok, Kwajalein, Dateline Time */
+    map.insert(40, QLatin1String("Pacific/Fiji"));              /* GMT+12:00 Fušál, Kamchatka, Mashall Is. */
+    map.insert(41, QLatin1String("Pacific/Noumea"));            /* GMT+11:00 Magadan, Solomon Is., New Caledonia */
+    map.insert(42, QLatin1String("Australia/Hobart"));          /* GMT+10:00 Hobart, Tasmania */
+    map.insert(43, QLatin1String("Pacific/Guam"));              /* GMT+10:00 Guam, Port Moresby */
+    map.insert(44, QLatin1String("Australia/Darwin"));          /* GMT+09:30 Darwin */
+    map.insert(45, QLatin1String("Asia/Shanghai"));             /* GMT+08:00 Beijing, Chongqing, Hong Kong SAR, Urumqi */
+    map.insert(46, QLatin1String("Asia/Omsk"));                 /* GMT+06:00 Almaty, Novosibirsk, North Central Asia */
+    map.insert(47, QLatin1String("Asia/Karachi"));              /* GMT+05:00 Islamabad, Karachi, Tashkent */
+    map.insert(48, QLatin1String("Asia/Kabul"));                /* GMT+04:30 Kabul */
+    map.insert(49, QLatin1String("Africa/Cairo"));              /* GMT+02:00 Cairo */
+    map.insert(50, QLatin1String("Africa/Harare"));             /* GMT+02:00 Harare, Pretoria */
+    map.insert(51, QLatin1String("Europe/Moscow"));             /* GMT+03:00 Moscow, St. Petersburg, Volgograd */
+    map.insert(53, QLatin1String("Atlantic/Cape_Verde"));       /* GMT-01:00 Cape Verde Is. */
+    map.insert(54, QLatin1String("Asia/Tbilisi"));              /* GMT+04:00 Baku, Tbilisi, Yerevan */
+    map.insert(55, QLatin1String("America/Tegucigalpa"));       /* GMT-06:00 Central America */
+    map.insert(56, QLatin1String("Africa/Nairobi"));            /* GMT+03:00 East Africa, Nairobi */
+    map.insert(58, QLatin1String("Asia/Yekaterinburg"));        /* GMT+05:00 Ekaterinburg */
+    map.insert(59, QLatin1String("Europe/Helsinki"));           /* GMT+02:00 Helsinki, Riga, Tallinn */
+    map.insert(60, QLatin1String("America/Greenland"));         /* GMT-03:00 Greenland */
+    map.insert(61, QLatin1String("Asia/Rangoon"));              /* GMT+06:30 Yangon (Rangoon) */
+    map.insert(62, QLatin1String("Asia/Katmandu"));             /* GMT+05:45 Kathmandu, Nepal */
+    map.insert(63, QLatin1String("Asia/Irkutsk"));              /* GMT+08:00 Irkutsk, Ulaan Bataar */
+    map.insert(64, QLatin1String("Asia/Krasnoyarsk"));          /* GMT+07:00 Krasnoyarsk */
+    map.insert(65, QLatin1String("America/Santiago"));          /* GMT-04:00 Santiago */
+    map.insert(66, QLatin1String("Asia/Colombo"));              /* GMT+06:00 Sri Jayawardenepura, Sri Lanka */
+    map.insert(67, QLatin1String("Pacific/Tongatapu"));         /* GMT+13:00 Nuku'alofa, Tonga */
+    map.insert(68, QLatin1String("Asia/Vladivostok"));          /* GMT+10:00 Vladivostok */
+    map.insert(69, QLatin1String("Africa/Bangui"));             /* GMT+01:00 West Central Africa */
+    map.insert(70, QLatin1String("Asia/Yakutsk"));              /* GMT+09:00 Yakutsk */
+    map.insert(71, QLatin1String("Asia/Dhaka"));                /* GMT+06:00 Astana, Dhaka */
+    map.insert(72, QLatin1String("Asia/Seoul"));                /* GMT+09:00 Seoul, Korea Standard time */
+    map.insert(73, QLatin1String("Australia/Perth"));           /* GMT+08:00 Perth, Western Australia */
+    map.insert(74, QLatin1String("Asia/Kuwait"));               /* GMT+03:00 Arab, Kuwait, Riyadh */
+    map.insert(75, QLatin1String("Asia/Taipei"));               /* GMT+08:00 Taipei */
+    map.insert(76, QLatin1String("Australia/Sydney"));          /* GMT+10:00 Canberra, Melbourne, Sydney */
+
+    return map;
+}
+
+static const QMap<int, QString> MSCDOTZIDTable = initMSCDOTZIDTable();
+
+QString Services::CalendarPrivate::checkAndConverCDOTZID(const QString& tzid, const Objects::Event* event)
+{
+    /* Try to match the @tzid to any valid timezone we know. */
+    KTimeZones timeZones;
+    const KTimeZones::ZoneMap zones = timeZones.zones();
+    KTimeZones::ZoneMap::const_iterator iter;
+    for (iter = zones.constBegin(); iter != zones.constEnd(); iter++) {
+        if (iter.key() == tzid) {
+            /* Yay, @tzid is a valid TZID in Olson format */
+            return tzid;
+        }
+    }
+
+    /* Damn, no match. Parse the iCal and try to find X-MICROSOFT-CDO-TZID
+     * property that we can match against the MSCDOTZIDTable */
+    ICalFormat format;
+    /* Use a copy of @event, otherwise it would be deleted when ptr is destroyed */
+    Incidence::Ptr ptr(new Event(*dynamic_cast<const Event*>(event)));
+    const QString vcard = format.toICalString(ptr);
+    const QStringList properties = vcard.split(QLatin1Char('\n'));
+    int CDOId = -1;
+    Q_FOREACH(const QString &property, properties) {
+        if (property.startsWith(QLatin1String("X-MICROSOFT-CDO-TZID"))) {
+            QStringList parsed = property.split(QLatin1Char(':'));
+            if (parsed.length() != 2) {
+                /* Fail */
+                return tzid;
+            }
+
+            CDOId = parsed.at(1).toInt();
+            break;
+        }
+    }
+
+    /* Wheeee, we have X-MICROSOFT-CDO-TZID, try to map it to Olson format */
+    if (CDOId > -1) {
+
+        /* *sigh* Some expert in MS assigned the same ID to two two different timezones... */
+        if (CDOId == 2) {
+
+            /* GMT Greenwich Mean Time: Dublin, Edinburgh, Lisbon, London */
+            if (tzid.contains(QLatin1String("Dublin")) ||
+                tzid.contains(QLatin1String("Edinburgh")) ||
+                tzid.contains(QLatin1String("Lisbon")) ||
+                tzid.contains(QLatin1String("London")))
+            {
+                return QLatin1String("Europe/London");
+            }
+
+            /* GMT+01:00 Sarajevo, Skopje, Sofija, Vilnius, Warsaw, Zagreb */
+            else if (tzid.contains(QLatin1String("Sarajevo")) ||
+                     tzid.contains(QLatin1String("Skopje")) ||
+                     tzid.contains(QLatin1String("Sofija")) ||
+                     tzid.contains(QLatin1String("Vilnius")) ||
+                     tzid.contains(QLatin1String("Warsaw")) ||
+                     tzid.contains(QLatin1String("Zagreb")))
+            {
+                return QLatin1String("Europe/Sarajevo");
+            }
+
+            /* This should never ever happen. I don't know what to do if it happens */
+            else {
+                return tzid;
+            }
+        }
+
+        if (MSCDOTZIDTable.contains(CDOId)) {
+            return MSCDOTZIDTable.value(CDOId);
+        }
+    }
+
+    /* Fail. The event does not have a valid timezone, does not contain
+     * X-MICROSOFT-CDO-TZID or we failed to map the CDO-TZID to a real timezone.
+     * We are screwed. Just return the original TZID and hope Google will accept it
+     * (though we know they won't) */
+    return tzid;
+}
